#include "iodna.h"

std::ostream& seqdna::display(int pos, int length, std::ostream& os){
  char lookup[4] = {'a','t','g','c'};
  
  for(int i=pos ; i < pos+length ; i++){
    int b    = i/4;
    int B    = i%4;
    int code = (data[b] >> (6 - 2*B)) & 0x3;
    /*
      #ifdef MASKED
      vector<int>           maskpos;
      #endif
    */
    os << lookup[code];
  }
  return os;
}


std::ostream& operator<<(std::ostream& os, seqdna& seq) {
  os << seq.display(0,seq.dnasize(),os); 
  return os;
}

/*-------------------*/

iodna::iodna() {;}

iodna::iodna(const string filename,bool concatseq){
  this->addFile(filename,concatseq);
}

iodna::~iodna(){ //FIXME -- erase each sequence inside _sequence
}


#ifdef MASKED	
  	
#define PUSH(nucleocode,nucleomask) {                               \
      code  |= (unsigned char) (nucleocode) << (6-(seq.bsize&0x7)); \
      if (nucleomask != mask) {                                     \
	mask = nucleomask;                                          \
        seq.maskpos.push_back(seq.bsize>>1);                        \
      }                                                             \
      seq.bsize += 2;                                               \
      if(!(seq.bsize&0x7)){                                         \
	seq.data.push_back(code);                                   \
        code = 0;                                                   \
      }                                                             \
}

#else
	
#define PUSH(nucleocode,nucleomask) {                               \
      code  |= (unsigned char) (nucleocode) << (6-(seq.bsize&0x7)); \
      seq.bsize += 2;                                               \
      if(!(seq.bsize&0x7)){                                         \
	seq.data.push_back(code);                                   \
        code = 0;                                                   \
      }                                                             \
}

#endif

/*
 * add file to the common data set
 * - concatseq (if you need to concatenate all multifasta in one data file)
 * - addmask   (create for each "seqdna" object an inner vector.
 */

/*
 *      A --> adenosine           M --> A C (amino)
 *      C --> cytidine            S --> G C (strong)
 *      G --> guanine             W --> A T (weak)
 *      T --> thymidine           B --> G T C
 *      U --> uridine             D --> G A T
 *      R --> G A (purine)        H --> A C T
 *      Y --> T C (pyrimidine)    V --> G C A
 *      K --> G T (keto)          N --> A G C T (any)
 */


int  iodna::addFile(const string filename, bool concatseq ){
  
  seqdna                seq ;
  char                  code = 0;
#ifdef MASKED
  char                  mask = 0;
#endif
  char                  strbuffer[256];

  // (a) initiate seq element
  seq.name    = "";
  seq.bsize   = 0;
  seq.data    = vector<unsigned char>();
#ifdef MASKED
  seq.maskpos = vector<int>();
#endif

  // (1) openning file
  ifstream ifs(filename.c_str());
  if (ifs <= 0){
    cerr << "* ERROR :  cant read \"" << filename << "\""  << endl;
    exit(-1);
  }
  
  // (2) scanning file
  while(!ifs.eof()){ 
    switch(ifs.get()){
      
    // defined code
    case 'A': 
      PUSH(0,0);
      break;
    case 'a':
      PUSH(0,1);
      break;

    case 'T': 
    case 'U':
      PUSH(1,0);
      break;
    case 't':
    case 'u':
      PUSH(1,1);
      break;   

    case 'G': 
      PUSH(2,0);
      break;
    case 'g':
      PUSH(2,1);
      break;

    case 'C':
      PUSH(3,0);
      break;
    case 'c':
      PUSH(3,1);
      break;

      /* undefined or half defined  */
    case 'N':
    case 'R':
    case 'Y':
    case 'K':
    case 'W':
    case 'S':
    case 'B':
    case 'D':
    case 'H':
    case 'V':
#ifdef ACGTONLY
      goto mem;
#else
      PUSH(0,0);
#endif
      break;
      
    case 'n':
    case 'r':
    case 'y':
    case 'k':
    case 'w':
    case 's':
    case 'b':
    case 'd':
    case 'h':
    case 'v':
#ifdef ACGTONLY
      goto mem;
#else
      PUSH(0,1);
#endif
      break;
      
    case '>':
      /* read fasta header name */
      ifs.getline(strbuffer,256);
      mem :
	strbuffer[255] = '\0';
      /* consider if sequences have to be concatenated */
      if (!concatseq){

	// save the last sequence recorded
	if ( seq.data.size() > 0){
	  // push the last byte
	  PUSH(0,0);PUSH(0,0);PUSH(0,0); 
	  seq.bsize-=2*3;
	  //
	  _sequences.push_back(seq);
	  cerr << " - chunk : " << strbuffer << ", dna size : " << seq.dnasize() << endl;
	}
	
	// clear current elements
	seq.name    = strbuffer;
	seq.bsize   = 0;
	seq.data.clear();
#ifdef MASKED
	seq.maskpos.clear();
#endif
	code  = 0;
      }
      break;
    }//switch	
  }//while
  ifs.close();

  // save the last sequence recorded
  if ( seq.data.size() > 0){
    // push the last byte
    PUSH(0,0);PUSH(0,0);PUSH(0,0); 
    seq.bsize-=2*3;
    // 
    _sequences.push_back(seq);
    cerr << " - chunk : " << strbuffer << ", dna size : " << seq.dnasize() << endl;
    // clear current elements
    seq.bsize   = 0;
    seq.data.clear();
#ifdef MASKED
    seq.maskpos.clear();
#endif
  }
  return 0;
}



/*	
 * Write DNA file (with mask : lower/uppercase)
 */

void iodna::writeFile(string & filename){
  
  // (1) openning file
  ofstream ofs(filename.c_str());
  if (ofs <= 0){
    cerr << "* ERROR :  cant write \"" << filename << "\""  << endl;
    exit(-1);
  }

  // (2) writing data
  ofs << (*this);
  ofs.close();
}



/*
 * Output current DNA data (with mask : lower/uppercase)
 */

std::ostream& operator<<(std::ostream& os, iodna& io){
   // (1) writing data
  for( vector<seqdna>::iterator it_vs  = io._sequences.begin(); it_vs !=  io._sequences.end(); it_vs++){
    // (1.1) write fasta chunk header
    os << ">" << it_vs->name << endl;   
    // (1.2) write  fasta chunk data
    int j = 0; 
    
#ifdef MASKED	
    int mask = 0;
    vector<int>::iterator it_m = it_vs->maskpos.begin();
#endif
    for(vector<unsigned char>::iterator it_v = it_vs->data.begin(); it_v != it_vs->data.end(); it_v++){
      // write nucleodite data
#ifdef MASKED
      // switch mask
      while (((*(it_m))>>2) == (j>>2)) {
	(++mask) %= 2;
	it_m++;
      }
      nucleotides(os,(*it_v)/* code */, (it_vs->dnasize() - j)/*remain*/, (mask)/*mask*/);
#else
      nucleotides(os,(*it_v)/* code */, (it_vs->dnasize() - j)/*remain*/, (0)/*mask*/);
#endif
      // write a carriage return
      j+= 4;
      if (!((j)%60))
	os << endl;
    }
    os << endl;
  }
  return os;
}


unsigned iodna::dnasize() {
  unsigned dnasize = 0;
  for( vector<seqdna>::iterator it_vs  = _sequences.begin(); it_vs !=  _sequences.end(); it_vs++){
    dnasize += it_vs->dnasize();
  }
  return dnasize;
}
  
unsigned iodna::bitsize() {
  unsigned bitsize = 0;
  for( vector<seqdna>::iterator it_vs  = _sequences.begin(); it_vs !=  _sequences.end(); it_vs++){
    bitsize += it_vs->bitsize();
  }
  return bitsize;
}
  

unsigned iodna::size() {
  unsigned size = 0;
  for( vector<seqdna>::iterator it_vs  = _sequences.begin(); it_vs !=  _sequences.end(); it_vs++){
    size += it_vs->size();
  }
  return size;
}
  
